/*____________________________________________________________________________
 *
 * vcif - program for validating a Crystallographic Information File
 *
 * Copyright (c) 1995-2004 International Union of Crystallography
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *
 * Author: Brian McMahon, IUCr, 5 Abbey Square, Chester CH1 2HU, England
 * E-mail: bm@iucr.org
____________________________________________________________________________*/

/*
The Crystallographic Information File (CIF) has been developed by the IUCr
as a standard for the electronic archival and transmission of
crystallographic data.  A CIF may contain numeric, character or textual
data, and is thus suited to the purpose of transmitting the text of a
paper reporting a crystallographic structure determination to IUCr
journals, primarily Acta Crystallographica, Section C. This program checks
the syntax of a file against the standard syntax rules for STAR files and
against the more restrictive CIF rules.

The Crystallographic Information File is described in the paper "The
Crystallographic Information File (CIF): a New Standard Archive File for
Crystallography" by S. R. Hall, F. H. Allen & I. D. Brown [Acta
Cryst. (1991), A45, 655-685], copies of which may be obtained from the
Technical Editor, IUCr, 5 Abbey Square, Chester CH1 2HU, England.

The STAR file is described by S. R. Hall
[J. Chem. Inf. Comput. Sci. (1991), 31, 326-333] and by S. R. Hall &
N. Spadaccini [J. Chem. Inf. Comput. Sci.  (1994), 34, 505-508].

The STAR concept is protected by UK Patent No. GB 2 243 467 B.

Copyright in the CIF specifications and Dictionary resides with the IUCr.
*/

#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "pass1.h"

/*
______________________________________________________________________________

Data definitions for this file
______________________________________________________________________________
*/

	extern int debug;
	extern int maxline;
	extern int maxdname;
	extern int reportlevel;
	extern int full_star;
	extern int printdblkname;
	extern long int errors_reported;
	extern char * dbgfile;
	extern char * line;

	char * buffer;
	char * bitof();
	char * strdup();
	char * strtok();

	void check_dataname();
	void check_dblkname();
	void check_savename();

	int state = BEGINNING_OF_FILE;
	int type;
	int lineno = 0;
	int packetsize = 0;
	int no_in_loop = 0;
	char message[256];
	char * current_dblock;
	char * current_save;
	FILE * debugfile;
	struct nlist *lookup();
	struct nlist *install();
	
	int no_globals = 0;	/* count no. of global_ blocks */
	char current_global[5];	/* string indexing current global_ scope */

/*
______________________________________________________________________________

The first pass through the file checks solely for breaches of STAR syntax and
the restrictive CIF rules. Each line is read into a buffer and tokenized into
atoms separated by white space, which are passed as STAR tokens, or
components of complex STAR tokens. This approach is not ideal from a
data-stream view of the file, but allows checking of line-width restrictions
and keeps track of line numbers for error reports.
______________________________________________________________________________
*/

void pass1(cif)
	FILE *cif;
{
#define BUFSIZE 4096

	/* void vcifwarn(); */
	int elaborate();


	int cifmaxline=2048;
	/* int maxline=80;  now declared in vcif.c */
	/* char * line; */
	/* char * buffer;  global in vcif.c */
	char * token;
	/* int type; */

	/* Open a file to log debugging info */
	if (debug == YES) {
		if ((debugfile = fopen(dbgfile ,"w")) == NULL) {
			fprintf(stderr, "vcif ERROR: Unable to open debug file - exiting\n");
			exit(3);
		}
	}


/* allocate space for current line (should have at most maxline characters) */
	line = (char *)malloc(maxline + 2);
/* allocate space for complete line (up to BUFSIZE) */
	buffer = (char *)malloc(BUFSIZE);
/* initialize current data block name to be null at start of file */
	current_dblock = strdup("");
/* initialize current save frame name to be null at start of file */
	current_save = strdup("");
/* initialize global flag */
	current_global[0] = 'g';
	current_global[1] = '0';
	current_global[2] = '\0';

	line[0] = '\0';
	buffer[0] = '\0';

/* main loop; read input file a line at a time */

	while (fgets(line, maxline+1, cif) != NULL) {

	/* This is a hand-coded and very simple finite-state machine. Token
	   types (which may be dependent on the current state) are returned
	   (and may change the state). Other actions may also be taken
	   (usually the generation of an error message if the new token is
	   out of context in the current state). */

		/* if the "line" array does not contain a newline, build up
		   a buffer of arbitrary size (up to BUFSIZE) */

		if (strchr(line, '\n') == NULL
			&& (feof(cif) == 0)) {	/* not end-of-file */

			if ((strlen(buffer) + strlen(line)) > BUFSIZE) {
				strncat(buffer, line, BUFSIZE-strlen(buffer));
			}
			else {
				strcat(buffer, line);
			}
			continue;
		}
		else  {
			if ((strlen(buffer)+strlen(line)) > BUFSIZE) {
				fprintf(stderr, "vcif WARNING: Line %d: length exceeds internal buffer size (%d) - rest discarded\n", lineno, BUFSIZE);
				strncat(buffer, line, BUFSIZE-strlen(buffer));
			}
			else {
				strcat(buffer, line);
				/* complain if line too long */
				if (strlen(buffer) > maxline + 1) {
					sprintf(message, "CIF ERROR: Line %d exceeds %d characters\n", lineno+1, maxline);
					vcifwarn(message, 3);
				}
			}
		lineno++;	/* increment line counter */
		}

		/* read the first token from the buffer */
		if ((token = (strtok(buffer, " \t\n\012\013\014\015"))) != NULL) {

			/* some states are invalid at start of line */
			switch (state){
				case BUILDING_SQUOT:
					sprintf(message, "ERROR: Single-quoted character string does not terminate at line %d\n", lineno-1);
					vcifwarn(message, 6);
					state = EXPECTING_DNAME;
					break;
				case IN_LOOP_BUILDING_SQUOT:
					sprintf(message, "ERROR: Single-quoted character string does not terminate at line %d\n", lineno-1);
					vcifwarn(message, 6);
					state = IN_LOOP_EXPECTING_VALUE;
					break;
				case BUILDING_DQUOT:
					sprintf(message, "ERROR: Double-quoted character string does not terminate at line %d\n", lineno-1);
					vcifwarn(message, 5);
					state = EXPECTING_DNAME;
					break;
				case IN_LOOP_BUILDING_DQUOT:
					sprintf(message, "ERROR: Double-quoted character string does not terminate at line %d\n", lineno-1);
					vcifwarn(message, 5);
					state = IN_LOOP_EXPECTING_VALUE;
					break;
				default:
					break;
			}


			/* If the first character (not just the first
			   non-white character) is a semicolon, must be
			   a text delimiter */
			if (token[0] == ';' && buffer[0] == token[0]) {
				switch (state){
					case BEGINNING_OF_FILE:
					case EXPECTING_DNAME:
						sprintf(message, "ERROR: Unexpected text field beginning at line %d\n", lineno);
						vcifwarn(message, 10);
						type = BEGTEXT;
						state = BUILDING_TEXT;
						break;
					case BUILDING_TEXT:
						type = ENDTEXT;
						state = EXPECTING_DNAME;
						break;
					case IN_LOOP_BUILDING_TEXT:
						type = ENDTEXT;
						no_in_loop++;
						state = IN_LOOP_EXPECTING_VALUE;
						break;
					case IN_LOOP_EXPECTING_DNAME:
						if (type == LOOPCODE) {
							type = BEGTEXT;
							state = BUILDING_TEXT;
							sprintf(message, "ERROR: loop_ keyword not followed by dataname at line %d\n", lineno);
							vcifwarn(message, 24);
							break;
						}
						else {
							type = BEGTEXT;
							state = IN_LOOP_BUILDING_TEXT;
							break;
						}
					case IN_LOOP_EXPECTING_VALUE:
						type = BEGTEXT;
						state = IN_LOOP_BUILDING_TEXT;
						break;
					default:
						type = BEGTEXT;
						state = BUILDING_TEXT;
						break;
				}
			if (debug == YES) fprintf(debugfile, "Token <%s> is of type %d (state %d)\n", token, type, state);
			}

			/* Now find the type of the current token (if not
			   text delimiter) */
			else {
				type = get_type_of(token);
				if (debug == YES) fprintf(debugfile, "Token <%s> is of type %d (state %d)\n", token, type, state);
				if (type == COMMENT) strtok(NULL,"\n");
				if (type == DATANAME) check_dataname(token);
			}

			/* Now tokenize rest of buffer */
			while ((token = (strtok(NULL, " \t\n\012\013\014\015"))) != NULL) {
				type = get_type_of(token);
				if (debug == YES) fprintf(debugfile, "Token <%s> is of type %d (state %d)\n", token, type, state);
				if (type == COMMENT) strtok(NULL,"\n");
				if (type == DATANAME) check_dataname(token);
			}
		}
		buffer[0]='\0';
	}		/* END OF MAIN WHILE LOOP */
	free(buffer);

	/* some states are invalid at end of file */
	switch (state) {
			case EXPECTING_DNAME:
				if (type == GLOBAL) {
					sprintf(message, "WARNING: Null global_ block at end of file\n");
					vcifwarn(message, 0);
				}
				if (type == BLOCKCODE) {
					sprintf(message, "WARNING: Null datablock at end of file\n");
					vcifwarn(message, 17);
				}
				break;
		case BEGINNING_OF_FILE:
			vcifwarn("WARNING: Null file\n", 23);
			break;
		case EXPECTING_VALUE:
			vcifwarn("ERROR: Missing data value at end of file\n", 11);
			break;
		case IN_LOOP_BUILDING_SQUOT:
		case IN_LOOP_BUILDING_DQUOT:
		case BUILDING_SQUOT:
		case BUILDING_DQUOT:
			vcifwarn("ERROR: Quoted character string at end of file does not terminate\n", 22);
			break;
		case IN_LOOP_BUILDING_TEXT:
		case BUILDING_TEXT:
			vcifwarn("ERROR: Text field at end of file does not terminate\n", 22);
			break;
		case IN_LOOP_EXPECTING_DNAME:
			vcifwarn("ERROR: Loop takes no data values before end of file\n", 15);
			break;
		case IN_LOOP_EXPECTING_VALUE:
			if ((packetsize != 0) && ((no_in_loop % packetsize) != 0)) {
				sprintf(message,
				"ERROR: Number of loop elements not multiple of packetsize at line %d\n", lineno);
				vcifwarn(message, 21);
			}
			break;
	}
	
	if (full_star == YES && current_save[0] != '\0') { 
		sprintf(message, "STAR ERROR: save_ frame %s not terminated at end of file\n", current_save);
		vcifwarn(message, 0);
	}

	return;
}


/*
______________________________________________________________________________

get_type_of() returns an integer indexing the type of the current token
______________________________________________________________________________
*/

int get_type_of(token)
	char * token;
{
	int n = 0;
	unsigned char x;
	char * s;

	s = token;
	while ((x =*s++) != '\0') {
		if(!isprint(x)) {
			sprintf(message, "WARNING: Non-printable ASCII character \\%#o at line %d\n", x, lineno);
			vcifwarn(message, 20);
		}
	}

/* A datablock identifier terminates a loop or occurs after a single value
   (when expecting a dataname) */
	if (strncmp(token, "data_", 5) == 0) {
		if (current_save[0] != '\0') {
			sprintf(message, "STAR ERROR: save_ frame %s not correctly terminated at line %d\n", current_save, lineno);
			vcifwarn(message, 0);
			current_save[0] = '\0';
		}
		switch (state) {
			case EXPECTING_DNAME:
				if (type == GLOBAL) {
					sprintf(message, "WARNING: Null global_ block prior to line %d\n", lineno);
					vcifwarn(message, 0);
				}
				if (type == BLOCKCODE) {
					sprintf(message, "WARNING: Null datablock prior to line %d\n", lineno);
					vcifwarn(message, 17);
				}
				type = BLOCKCODE;
				break;
			case BUILDING_SQUOT:
			case BUILDING_DQUOT:
			case BUILDING_TEXT:
			case IN_LOOP_BUILDING_SQUOT:
			case IN_LOOP_BUILDING_DQUOT:
			case IN_LOOP_BUILDING_TEXT:
				type = CHAR;
				break;
			case BEGINNING_OF_FILE:
				state = EXPECTING_DNAME;
				type = BLOCKCODE;
				break;
			case IN_LOOP_EXPECTING_DNAME:
				sprintf(message, "ERROR: Loop takes no data values before data block code at line %d\n", lineno);
				vcifwarn(message, 15);
				state = EXPECTING_DNAME;
				type = BLOCKCODE;
				break;
			case EXPECTING_VALUE:
				sprintf(message, "ERROR: Missing data value before data block code at line %d\n", lineno);
				vcifwarn(message, 7);
				state = EXPECTING_DNAME;
				type = BLOCKCODE;
				break;
			case IN_LOOP_EXPECTING_VALUE:
				/* need to count no. of loop data values */
				state = EXPECTING_DNAME;
				type = BLOCKCODE;
				if ((packetsize != 0) && ((no_in_loop % packetsize) != 0)) {
					sprintf(message,
					"ERROR: Number of loop elements not multiple of packetsize at line %d\n", lineno);
					vcifwarn(message, 21);
				}
				packetsize = 0;		/* exiting from loop */
				break;
			default:
				type = BLOCKCODE;
				break;
		}

	/* check block code length and uniqueness within a file */
	if (type == BLOCKCODE) check_dblkname(token);

	return type;
	}

/* A global_ keyword in the full STAR syntax can occur wherever a data_
   blockcode can occur */
	if (strcmp(token, "global_") == 0 && full_star == YES) {
		switch (state) {
			case EXPECTING_DNAME:
				if (type == BLOCKCODE) {
					sprintf(message, "WARNING: Null datablock prior to line %d\n", lineno);
					vcifwarn(message, 17);
				}
				if (type == GLOBAL) {
					sprintf(message, "WARNING: Null global_ block prior to line %d\n", lineno);
					vcifwarn(message, 0);
				}
				type = GLOBAL;
				break;
			case BUILDING_SQUOT:
			case BUILDING_DQUOT:
			case BUILDING_TEXT:
			case IN_LOOP_BUILDING_SQUOT:
			case IN_LOOP_BUILDING_DQUOT:
			case IN_LOOP_BUILDING_TEXT:
				type = CHAR;
				break;
			case BEGINNING_OF_FILE:
				state = EXPECTING_DNAME;
				type = GLOBAL;
				break;
			case IN_LOOP_EXPECTING_DNAME:
				sprintf(message, "ERROR: Loop takes no data values before data block code at line %d\n", lineno);
				vcifwarn(message, 15);
				state = EXPECTING_DNAME;
				type = GLOBAL;
				break;
			case EXPECTING_VALUE:
				sprintf(message, "ERROR: Missing data value before data block code at line %d\n", lineno);
				vcifwarn(message, 7);
				state = EXPECTING_DNAME;
				type = GLOBAL;
				break;
			case IN_LOOP_EXPECTING_VALUE:
				/* need to count no. of loop data values */
				state = EXPECTING_DNAME;
				type = GLOBAL;
				if ((packetsize != 0) && ((no_in_loop % packetsize) != 0)) {
					sprintf(message,
					"ERROR: Number of loop elements not multiple of packetsize at line %d\n", lineno);
					vcifwarn(message, 21);
				}
				packetsize = 0;		/* exiting from loop */
				break;
			default:
				type = GLOBAL;
				break;
		}

	/* Increment global_ counter and generate new scoping flag */
	if (type == GLOBAL) {
		no_globals++;
		sprintf(current_global, "g%d", no_globals);
	}

	return type;
	}

/* The save_ keyword in the full STAR syntax terminates a named save_
   frame */
	if (strcmp(token, "save_") == 0 && full_star == YES) {
		switch (state) {
			case EXPECTING_DNAME:
				if (current_save[0] == '\0') {
					sprintf(message, "STAR ERROR: Misplaced save_ keyword at line %d\n", lineno);
					vcifwarn(message, 0);
				}
				if (type == SAVE_BEGIN) {
					sprintf(message, "WARNING: Null save_ frame ending at line %d\n", lineno);
					vcifwarn(message, 0);
				}
				type = SAVE_END;
				current_save[0] = '\0';
				break;
			case BUILDING_SQUOT:
			case BUILDING_DQUOT:
			case BUILDING_TEXT:
			case IN_LOOP_BUILDING_SQUOT:
			case IN_LOOP_BUILDING_DQUOT:
			case IN_LOOP_BUILDING_TEXT:
				type = CHAR;
				break;
			case BEGINNING_OF_FILE:
				sprintf(message, "STAR ERROR: save_ keyword at beginning of file\n");
				vcifwarn(message, 0);
				state = EXPECTING_DNAME;
				type = SAVE_END;
				break;
			case IN_LOOP_EXPECTING_DNAME:
				sprintf(message, "STAR ERROR: Loop takes no data values before save_ keyword at line %d\n", lineno);
				vcifwarn(message, 0);
				state = EXPECTING_DNAME;
				if (current_save[0] == '\0') {
					sprintf(message, "STAR ERROR: Misplaced save_ keyword at line %d\n", lineno);
					vcifwarn(message, 0);
				}
				if (type == SAVE_BEGIN) {
					sprintf(message, "WARNING: Null save_ frame ending at line %d\n", lineno);
					vcifwarn(message, 0);
				}
				type = SAVE_END;
				current_save[0] = '\0';
				break;
			case EXPECTING_VALUE:
				sprintf(message, "STAR ERROR: Missing data value before save_ keyword at line %d\n", lineno);
				vcifwarn(message, 0);
				state = EXPECTING_DNAME;
				if (current_save[0] == '\0') {
					sprintf(message, "STAR ERROR: Misplaced save_ keyword at line %d\n", lineno);
					vcifwarn(message, 0);
				}
				if (type == SAVE_BEGIN) {
					sprintf(message, "WARNING: Null save_ frame ending at line %d\n", lineno);
					vcifwarn(message, 0);
				}
				type = SAVE_END;
				current_save[0] = '\0';
				break;
			case IN_LOOP_EXPECTING_VALUE:
				/* need to count no. of loop data values */
				state = EXPECTING_DNAME;
				if (current_save[0] == '\0') {
					sprintf(message, "STAR ERROR: Misplaced save_ keyword at line %d\n", lineno);
					vcifwarn(message, 0);
				}
				if (type == SAVE_BEGIN) {
					sprintf(message, "WARNING: Null save_ frame ending at line %d\n", lineno);
					vcifwarn(message, 0);
				}
				type = SAVE_END;
				current_save[0] = '\0';
				if ((packetsize != 0) && ((no_in_loop % packetsize) != 0)) {
					sprintf(message,
					"ERROR: Number of loop elements not multiple of packetsize at line %d\n", lineno);
					vcifwarn(message, 21);
				}
				packetsize = 0;		/* exiting from loop */
				break;
			default:
				if (current_save[0] == '\0') {
					sprintf(message, "STAR ERROR: Misplaced save_ keyword at line %d\n", lineno);
					vcifwarn(message, 0);
				}
				if (type == SAVE_BEGIN) {
					sprintf(message, "WARNING: Null save_ frame ending at line %d\n", lineno);
					vcifwarn(message, 0);
				}
				type = SAVE_END;
				current_save[0] = '\0';
				break;
		}
	return type;
	}

/* A save_frame declaration in the full STAR syntax can occcur where a
   data_ block identifier can */
	if (strncmp(token, "save_", 5) == 0 && full_star == YES) {
		if (current_save[0] != '\0') {
			sprintf(message, "STAR ERROR: save_ frame %s not correctly terminated at line %d\n", current_save, lineno);
			vcifwarn(message, 0);
			current_save[0] = '\0';
		}
		switch (state) {
			case EXPECTING_DNAME:
				type = SAVE_BEGIN;
				break;
			case BUILDING_SQUOT:
			case BUILDING_DQUOT:
			case BUILDING_TEXT:
			case IN_LOOP_BUILDING_SQUOT:
			case IN_LOOP_BUILDING_DQUOT:
			case IN_LOOP_BUILDING_TEXT:
				type = CHAR;
				break;
			case BEGINNING_OF_FILE:
				sprintf(message, "STAR ERROR: save_ frame at beginning of file\n");
				vcifwarn(message, 0);
				state = EXPECTING_DNAME;
				type = SAVE_BEGIN;
				break;
			case IN_LOOP_EXPECTING_DNAME:
				sprintf(message, "ERROR: Loop takes no data values before save frame at line %d\n", lineno);
				vcifwarn(message, 0);
				state = EXPECTING_DNAME;
				type = SAVE_BEGIN;
				break;
			case EXPECTING_VALUE:
				sprintf(message, "ERROR: Missing data value before save frame at line %d\n", lineno);
				vcifwarn(message, 0);
				state = EXPECTING_DNAME;
				type = SAVE_BEGIN;
				break;
			case IN_LOOP_EXPECTING_VALUE:
				/* need to count no. of loop data values */
				state = EXPECTING_DNAME;
				type = SAVE_BEGIN;
				if ((packetsize != 0) && ((no_in_loop % packetsize) != 0)) {
					sprintf(message,
					"ERROR: Number of loop elements not multiple of packetsize at line %d\n", lineno);
					vcifwarn(message, 21);
				}
				packetsize = 0;		/* exiting from loop */
				break;
			default:
				type = SAVE_BEGIN;
				break;
		}

	/* check frame code length and uniqueness within file scope */
	if (type == SAVE_BEGIN) check_savename(token);

	return type;
	}

/* The "loop_" keyword prefixes a cluster of datanames followed by associated
   values */
	if (strcmp(token, "loop_") == 0) {
		switch (state) {
			case BUILDING_SQUOT:
			case BUILDING_DQUOT:
			case BUILDING_TEXT:
			case IN_LOOP_BUILDING_SQUOT:
			case IN_LOOP_BUILDING_DQUOT:
			case IN_LOOP_BUILDING_TEXT:
				type = CHAR;
				break;
			case EXPECTING_DNAME:
				type = LOOPCODE;
				state = IN_LOOP_EXPECTING_DNAME;
				break;
			case IN_LOOP_EXPECTING_DNAME:
				if (type == LOOPCODE) {
					sprintf(message, "ERROR: Redundant loop_ keyword at line %d\n", lineno);
					vcifwarn(message, 9);
				}
				else {
					sprintf(message, "ERROR: Unexpected loop_ keyword at line %d\n", lineno);
					vcifwarn(message, 16);
				}
				type = LOOPCODE;
				state = IN_LOOP_EXPECTING_DNAME;
				break;
			case IN_LOOP_EXPECTING_VALUE:
				if ((packetsize != 0) && ((no_in_loop % packetsize) != 0)) {
					sprintf(message,
					"ERROR: Number of loop elements not multiple of packetsize at line %d\n", lineno);
					vcifwarn(message, 21);
				}
				packetsize = 0;		/* exiting from loop */
				type = LOOPCODE;
				state = IN_LOOP_EXPECTING_DNAME;
				break;
			default:
				type = LOOPCODE;
				sprintf(message, "ERROR: loop_ keyword out of context at line %d\n", lineno);
				vcifwarn(message, 16);
				state = IN_LOOP_EXPECTING_DNAME;
				break;
		}
	if (type == LOOPCODE) { packetsize = 0; no_in_loop = 0; }
	return type;
	}

/* The "#" character (if not within a character string) opens a
   comment to the end of the current line */
	if (token[0] == '#') {
		switch (state) {
			case BUILDING_SQUOT:
				if (token[strlen(token) - 1] == '\'') {
					type = ENDSQUOT;
					state = EXPECTING_DNAME;
					break;
				}
			case BUILDING_DQUOT:
				if (token[strlen(token) - 1] == '\"') {
					type = ENDSQUOT;
					state = EXPECTING_DNAME;
					break;
				}
			case BUILDING_TEXT:
			case IN_LOOP_BUILDING_SQUOT:
				if (token[strlen(token) - 1] == '\"') {
					state = IN_LOOP_EXPECTING_VALUE;
					type = ENDDQUOT;
						no_in_loop++;
					break;
				}
			case IN_LOOP_BUILDING_DQUOT:
				if (token[strlen(token) - 1] == '\"') {
					state = IN_LOOP_EXPECTING_VALUE;
					type = ENDDQUOT;
						no_in_loop++;
					break;
				}
			case IN_LOOP_BUILDING_TEXT:
				type = CHAR;
				break;
			default:
				type = COMMENT;
				break;
		}
	return type;
	}

/* Leading underscore indicates a dataname */
	if (token[0] == '_') {
		switch (state) {
			case BUILDING_SQUOT:
				if (token[strlen(token) - 1] == '\'') {
					type = ENDSQUOT;
					state = EXPECTING_DNAME;
					break;
				}
			case BUILDING_DQUOT:
				if (token[strlen(token) - 1] == '\"') {
					type = ENDSQUOT;
					state = EXPECTING_DNAME;
					break;
				}
			/* Not sure why this was here - moved down to
			   case IN_LOOP_BUILDING_TEXT (BM:1998-08-06)
			   case BUILDING_TEXT: */
			case IN_LOOP_BUILDING_SQUOT:
				if (token[strlen(token) - 1] == '\'') {
					state = IN_LOOP_EXPECTING_VALUE;
					type = ENDSQUOT;
						no_in_loop++;
					break;
				}
			case IN_LOOP_BUILDING_DQUOT:
				if (token[strlen(token) - 1] == '\"') {
					state = IN_LOOP_EXPECTING_VALUE;
					type = ENDDQUOT;
						no_in_loop++;
					break;
				}
			case BUILDING_TEXT:
			case IN_LOOP_BUILDING_TEXT:
				type = CHAR;
				break;
			case BEGINNING_OF_FILE:
				if (strlen(token) == 1) { /* pathological case */
					sprintf(message, "ERROR: Single underscore ('_') at line %d not defined\n", lineno);
					vcifwarn(message, 18);
					type = UNDEFINED;
				}
				else {
					sprintf(message, "ERROR: No data block code before dataname at line %d\n", lineno);
					vcifwarn(message, 8);
					type = DATANAME;
					state = EXPECTING_VALUE;
				}
				break;
			case EXPECTING_VALUE:
				if (strlen(token) == 1) { /* pathological case */
					sprintf(message, "ERROR: Single underscore ('_') at line %d not defined\n", lineno);
					vcifwarn(message, 18);
					type = UNDEFINED;
				}
				else {
					sprintf(message, "ERROR: Value missing at line %d\n", lineno);
					vcifwarn(message, 11);
					type = DATANAME;
				}
				break;
			case IN_LOOP_EXPECTING_DNAME:
				if (strlen(token) == 1) { /* pathological case */
					sprintf(message, "ERROR: Single underscore ('_') at line %d not defined\n", lineno);
					vcifwarn(message, 18);
					type = UNDEFINED;
				}
				else {
					state = IN_LOOP_EXPECTING_DNAME;
					type = DATANAME;
				}
				packetsize++;
				break;
			case IN_LOOP_EXPECTING_VALUE:
				if (strlen(token) == 1) { /* pathological case */
					sprintf(message, "ERROR: Single underscore ('_') at line %d not defined\n", lineno);
					vcifwarn(message, 18);
					type = UNDEFINED;
				}
				else {
					state = EXPECTING_VALUE;
					type = DATANAME;
				}
				if ((packetsize != 0) && ((no_in_loop % packetsize) != 0)) {
					sprintf(message,
					"ERROR: Number of loop elements not multiple of packetsize at line %d\n", lineno);
					vcifwarn(message, 21);
				}
				packetsize = 0;		/* exiting from loop */
				break;
			default:
				if (strlen(token) == 1) { /* pathological case */
					sprintf(message, "ERROR: Single underscore ('_') at line %d not defined\n", lineno);
					vcifwarn(message, 18);
					type = UNDEFINED;
				}
				else {
					type = DATANAME;
					state = EXPECTING_VALUE;
				}
				break;
		}
	return type;
	}

/* Messy stuff: leading quote opens char string that may include white space */
	if (token[0] == '\'' && (strlen(token) == 1 || token[strlen(token)-1] != '\'')) {
		switch (state) {
			case IN_LOOP_BUILDING_SQUOT:
				if (strlen(token) == 1) {
					type = ENDSQUOT;
						no_in_loop++;
					state = IN_LOOP_EXPECTING_VALUE;
					break;
				}
			case BUILDING_SQUOT:
				if (strlen(token) == 1) {
					type = ENDSQUOT;
					state = EXPECTING_DNAME;
					break;
				}
			case IN_LOOP_BUILDING_DQUOT:
				if (token[strlen(token) - 1] == '\"') {
					state = IN_LOOP_EXPECTING_VALUE;
					type = ENDDQUOT;
						no_in_loop++;
					break;
				}
			case IN_LOOP_BUILDING_TEXT:
			case BUILDING_DQUOT:
				if (token[strlen(token) - 1] == '\"') {
					state = EXPECTING_DNAME;
					type = ENDDQUOT;
					break;
				}
			case BUILDING_TEXT:
				type = CHAR;
				break;
			case BEGINNING_OF_FILE:
			case EXPECTING_DNAME:
				sprintf(message, "ERROR: Unexpected string (%s) at line %d\n", (char *)bitof(token), lineno);
				vcifwarn(message, 10);
				/* state = EXPECTING_DNAME; */
				break;
			case EXPECTING_VALUE:
				state = BUILDING_SQUOT;
				type = BEGSQUOT;
				break;
			case IN_LOOP_EXPECTING_DNAME:
				if (type == LOOPCODE) {
					type = BEGSQUOT;
					state = BUILDING_SQUOT;
					sprintf(message, "ERROR: loop_ keyword not followed by dataname at line %d\n", lineno);
					vcifwarn(message, 24);
					break;
				}	/* else fall through */
			case IN_LOOP_EXPECTING_VALUE:
				state = IN_LOOP_BUILDING_SQUOT;
				type = BEGSQUOT;
				break;
		}
	return type;
	}

/* Messy stuff: leading quote opens char string that may include white space */
	if (token[0] == '\"' && (strlen(token) == 1 || token[strlen(token)-1] != '\"')) {
		switch (state) {
			case IN_LOOP_BUILDING_DQUOT:
				if (strlen(token) == 1) {
					type = ENDDQUOT;
						no_in_loop++;
					state = IN_LOOP_EXPECTING_VALUE;
					break;
				}
			case BUILDING_DQUOT:
				if (strlen(token) == 1) {
					type = ENDDQUOT;
					state = EXPECTING_DNAME;
					break;
				}
			case IN_LOOP_BUILDING_SQUOT:
				if (token[strlen(token) - 1] == '\'') {
					state = IN_LOOP_EXPECTING_VALUE;
					type = ENDSQUOT;
						no_in_loop++;
					break;
				}
			case IN_LOOP_BUILDING_TEXT:
			case BUILDING_SQUOT:
				if (token[strlen(token) - 1] == '\'') {
					state = EXPECTING_DNAME;
					type = ENDSQUOT;
					break;
				}
			case BUILDING_TEXT:
				type = CHAR;
				break;
			case BEGINNING_OF_FILE:
			case EXPECTING_DNAME:
				sprintf(message, "ERROR: Unexpected string (%s) at line %d\n", (char *)bitof(token), lineno);
				vcifwarn(message, 10);
				break;
			case EXPECTING_VALUE:
				state = BUILDING_DQUOT;
				type = BEGDQUOT;
				break;
			case IN_LOOP_EXPECTING_DNAME:
				if (type == LOOPCODE) {
					type = BEGDQUOT;
					state = BUILDING_DQUOT;
					sprintf(message, "ERROR: loop_ keyword not followed by dataname at line %d\n", lineno);
					vcifwarn(message, 24);
					break;
				}	/* else fall through */
			case IN_LOOP_EXPECTING_VALUE:
				state = IN_LOOP_BUILDING_DQUOT;
				type = BEGDQUOT;
				break;
		}
	return type;
	}

/* All other possibilities have been exhausted - default case */
		switch (state) {
			case BUILDING_SQUOT:
				if (token[strlen(token) - 1] == '\'') {
					state = EXPECTING_DNAME;
					type = ENDSQUOT;
				}
				break;
			case BUILDING_DQUOT:
				if (token[strlen(token) - 1] == '\"') {
					state = EXPECTING_DNAME;
					type = ENDDQUOT;
				}
				break;
			case IN_LOOP_BUILDING_SQUOT:
				if (token[strlen(token) - 1] == '\'') {
					state = IN_LOOP_EXPECTING_VALUE;
					type = ENDSQUOT;
						no_in_loop++;
				}
				break;
			case IN_LOOP_BUILDING_DQUOT:
				if (token[strlen(token) - 1] == '\"') {
					state = IN_LOOP_EXPECTING_VALUE;
					type = ENDDQUOT;
						no_in_loop++;
				}
				break;
			case BUILDING_TEXT:
			case IN_LOOP_BUILDING_TEXT:
				type = CHAR;
				break;
			case BEGINNING_OF_FILE:
				sprintf(message, "ERROR: File begins incorrectly with text (%s) at line %d\n", (char *)bitof(token), lineno);
				vcifwarn(message, 19);
				state = EXPECTING_DNAME;
				type = CHAR;
				break;
			case IN_LOOP_EXPECTING_DNAME:
				if (type == LOOPCODE) {
					type = CHAR;
					state = EXPECTING_DNAME;
					sprintf(message, "ERROR: loop_ keyword not followed by dataname at line %d\n", lineno);
					vcifwarn(message, 24);
					break;
				}	/* else fall through */
			case IN_LOOP_EXPECTING_VALUE:
				state = IN_LOOP_EXPECTING_VALUE;
				no_in_loop++;
		if (debug == YES) fprintf(debugfile, "Number in current loop of packet size %d = %d\n", packetsize, no_in_loop);
				type = CHAR;
				break;
			case EXPECTING_DNAME:
				sprintf(message, "ERROR: Unexpected string (%s) at line %d\n", (char *)bitof(token), lineno);
				vcifwarn(message, 10);
				state = EXPECTING_DNAME;
				type = CHAR;
				break;
			default:
				state = EXPECTING_DNAME;
				type = CHAR;
				break;
		}

	return type;
}

/*
______________________________________________________________________________

check_dataname() checks some data name attributes
______________________________________________________________________________
*/

void check_dataname(dname)
	char * dname;
{
	char * instance;

	if (strlen(dname) > maxdname) {
		/* allows user to redefine max. line length */
		if (strlen(dname) > CIFMAXDNAME) {
			sprintf(message, "CIF ERROR: Line %d: dataname exceeds %d characters (%s)\n", lineno, maxdname, bitof(dname));
			vcifwarn(message, 4);
		}
		else {
			sprintf(message, "WARNING: Line %d: dataname exceeds %d characters (%s)\n", lineno, maxdname, bitof(dname));
			vcifwarn(message, 2);
		}
	}

/* dataname must be unique in each datablock - store blockname:dataname */
	if ((strlen(current_dblock) + strlen(dname)) != 0)
		instance = (char *)malloc(strlen(current_global) + strlen(current_save) + strlen(current_dblock) + strlen(dname) + 5);
	strcpy(instance, current_global);
	strcat(instance, "\r");
	strcat(instance, current_save);
	strcat(instance, "\f");
	strcat(instance, current_dblock);
	strcat(instance, "\t");
	strcat(instance, dname); 

	if (debug == YES) fprintf(debugfile, "[<---- Current global:save:block:dataname is %s ---->]\n", instance);
	if (lookup(instance) == NULL) {
		install(instance);
	}
	else {
		sprintf(message, "ERROR: Duplicate dataname %s in data_%s at line %d\n", dname, bitof(current_dblock), lineno);
		vcifwarn(message, 14);
	}
	return;
}

/*
______________________________________________________________________________

check_dblkname() checks some block codes attributes
______________________________________________________________________________
*/

void check_dblkname(token)
	char * token;
{
	if (printdblkname == YES) printf("%s\n", token);
	if (strlen(token) > maxdname) {
		/* allows user to redefine max. length */
		if (strlen(token) > CIFMAXDNAME) {
			sprintf(message, "CIF ERROR: Line %d: dataname exceeds %d characters (%s)\n", lineno, maxdname, (char *)bitof(token));
			vcifwarn(message, 4);
		}
		else {
			sprintf(message, "WARNING: Line %d: dataname exceeds %d characters (%s)\n", lineno, maxdname, token);
			vcifwarn(message, 2);
		}
	}

	if (strlen(token) == 5) {
		/* "data_" without trailing string is invalid */
		sprintf(message, "ERROR: data_ declaration at line %d needs identifier\n", lineno);
		vcifwarn(message, 12);
	}
	else {
		/* block code must be unique in file */
		current_dblock = strdup(token+5);
		if (debug == YES) fprintf(debugfile, "[<---- Current dataname is data_%s ---->]\n", current_dblock);
		if (lookup(current_dblock) == NULL) {
			install(current_dblock);
		}
		else {
			sprintf(message, "ERROR: Duplicate datablock name data_%s at line %d\n", current_dblock, lineno);
			vcifwarn(message, 13);
		}
	}
}

/*
______________________________________________________________________________

check_savename() checks some block codes attributes
______________________________________________________________________________
*/

void check_savename(token)
	char * token;
{
	char * instance;

	current_save = strdup(token+5);

	/* frame code must be unique in current global_ and data block */
	instance = malloc(strlen(current_global) + strlen(current_save) + strlen(current_dblock) + 4);
	strcpy(instance, current_global);
	strcat(instance, "\r");
	strcat(instance, current_save);
	strcat(instance, "\f");
	strcat(instance, current_dblock);

	current_save = strdup(token+5);
	if (debug == YES) fprintf(debugfile, "[<---- Current saveframe is save_%s ---->]\n", current_save);
	if (debug == YES) fprintf(debugfile, "[<---- Current GSD instance is %s ---->]\n", instance);
	if (lookup(instance) == NULL) {
		install(instance);
	}
	else {
		sprintf(message, "STAR ERROR: Duplicate save frame name save_%s at line %d\n", current_save, lineno);
		vcifwarn(message, 0);
	}
	return;
}


/*
______________________________________________________________________________

Hash table stuff
______________________________________________________________________________
*/

struct nlist {
	struct nlist *next;
	char *name;
};

#define HASHSIZE 79

static struct nlist *hashtab[HASHSIZE];

unsigned int hash(s)
	char *s;
{
	unsigned hashval;

	for (hashval = 0; *s != '\0'; s++)
		hashval = *s + 31 * hashval;
	return hashval % HASHSIZE;
}

struct nlist *lookup(s)
	char *s;
{
	struct nlist *np;

	for (np = hashtab[hash(s)]; np != NULL; np = np->next)
		if (strcmp(s, np->name) == 0)
			return np;
	return NULL;
}



struct nlist *install(name)
 	char *name;
{
	struct nlist *np;
	unsigned hashval;

	if ((np = lookup(name)) == NULL) {
		np = (struct nlist *) malloc(sizeof(*np));
		if (np == NULL || (np->name = strdup(name)) == NULL)
			return NULL;
		hashval = hash(name);
		np->next = hashtab[hashval];
		hashtab[hashval] = np;
	}
	return np;
}

char *bitof(s)
char *s;
{
   int n = 54;
   int i = 0;
   static char p[54];

   while (*s != '\0' && i<n) {
     p[i]= *s;
     s++;
     if (i>50) p[i] = '.';
     i++;
   }
     p[i]= '\0';
return p;
}
