/*____________________________________________________________________________
 *
 * vcif - program for validating a Crystallographic Information File
 *
 * Copyright (c) 1995-2004 International Union of Crystallography
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *
 * Author: Brian McMahon, IUCr, 5 Abbey Square, Chester CH1 2HU, England
 * E-mail: bm@iucr.org
 *
 * Rev 1.2:  H. J. Bernstein, Dowling College, Oakdale, NY 11769, USA
 * E-mail: yaya@dowling.edu
____________________________________________________________________________*/

/*
The Crystallographic Information File (CIF) has been developed by the IUCr
as a standard for the electronic archival and transmission of
crystallographic data.  A CIF may contain numeric, character or textual
data, and is thus suited to the purpose of transmitting the text of a
paper reporting a crystallographic structure determination to IUCr
journals, primarily Acta Crystallographica, Section C. This program checks
the syntax of a file against the standard syntax rules for STAR files and
against the more restrictive CIF rules.

The Crystallographic Information File is described in the paper "The
Crystallographic Information File (CIF): a New Standard Archive File for
Crystallography" by S. R. Hall, F. H. Allen & I. D. Brown [Acta
Cryst. (1991), A45, 655-685], copies of which may be obtained from the
Technical Editor, IUCr, 5 Abbey Square, Chester CH1 2HU, England.

The STAR file is described by S. R. Hall
[J. Chem. Inf. Comput. Sci. (1991), 31, 326-333] and by S. R. Hall &
N. Spadaccini [J. Chem. Inf. Comput. Sci.  (1994), 34, 505-508].

The STAR concept is protected by UK Patent No. GB 2 243 467 B.

Copyright in the CIF specifications and Dictionary resides with the IUCr.
*/


/*
******************************************************************************

A general program description will be given here later.

******************************************************************************
*/


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "pass1.h"



/*
______________________________________________________________________________

Data definitions for this file
______________________________________________________________________________
*/

	FILE *cif;
	extern char * buffer;

#define QUIET 0
#define TERSE 1
#define NORMAL 2
#define VERBOSE 3
#define MAXERRS 256  /* because shell reports return status modulo 256 */

    /*  For the following variables, negative values or used for
	    defaults.  These will be reset to the appropriate values
		depending on the interaction with the user set values   */

	int maxline = -2048;    /* max. line length in input file    */
	int maxdname = -75;		/* max. size of block code/dataname  */
	int ciflevel = -1;      /* CIF level: 1 for 1.1, 0 for 1.0   */
	
	int ignoreargs;         /* flag to ignore further arguments  */
	
	
	int reportlevel = NORMAL;	/* controls verbosity of output  */
	long int errors_reported = 0;	/* counts no. of errors      */
	int error_threshold = -20;	/* no. reported in verbose mode  */
	int printdblkname = NO;		/* print data block names        */
	int force_threshold = NO;	/* apply threshold to quiet mode */
	int full_star = NO;		    /* allow full STAR syntax        */
	char *dbgfile;
	int debug = NO;
	char * line;
	extern int lineno;
	
/*  NUMERICARG is a macro for processing nermeric
	command line arguments
	
	xshort -- a quoted string such as "-L", giving the
	          short form of the command line argument.
			  (always precisely 2 characters).
    xlong  -- a quoted string shuch as "--level",
	          giving the long form of the same command
			  line argument.
    variable -- the variable of type int to hold the
	          value of the command line argument.
	name   -- a quoted string giving the meaning of
	          the variable.
    low    -- the lowest valid value of variable
	high   -- the highest valid value of variable
	                                                             */
#define NUMERICARG(xshort,xlong,variable,name,low,high)                              \
		else if (strncmp(arg, (xshort),2) == 0) {                                    \
		    tempval = (int)strtol(arg+2,&endptr,10);                                 \
			if (endptr==arg+2) tempval = (int)strtol((arg=av[++ax]),&endptr,10);     \
			if (*endptr != '\0') {                                                   \
			  errflg++;                                                              \
			} else {                                                                 \
			  if ( tempval < (low) || tempval > (high)) {                            \
			    errflg++;                                                            \
			    fprintf(stderr, "vcif ERROR: invalid %s\n",(name));                  \
			  } else {                                                               \
			  	(variable) = tempval;                                                \
			  }                                                                      \
			}                                                                        \
		}                                                                            \
		else if (strcmp(arg, (xlong)) == 0) {                                        \
		    tempval = (int)strtol(av[++ax],&endptr,10);                              \
			if (*endptr != '\0') {                                                   \
			  errflg++;                                                              \
			} else {                                                                 \
			  if (tempval < (low) || tempval > (high)) {                             \
			    errflg++;                                                            \
			    fprintf(stderr, "vcif ERROR: invalid %s\n",(name));                  \
			  } else {                                                               \
			  	(variable) = tempval;                                                \
			  }                                                                      \
			}                                                                        \
		}


/*
______________________________________________________________________________

main program segment processes command line and drives passes through file
______________________________________________________________________________
*/

int main(int ac, char *av[])
{
	int ax;
	char *arg;
	char *endptr=NULL;
	int errflg;
	int tempval;
	
	char *inputfile = "stdin";
	
	errflg = 0;

	ignoreargs = NO;

	for (ax=1; (ax<ac && !ignoreargs); ax++) {	/* crude processing of command-line */
		arg = av[ax];
		/* -Ddbgfile or -D dbgfile or --debug dbgfile */
		if (strncmp(arg, "-D",2) == 0) {
			debug = YES;
			dbgfile = arg+2;
			if (*dbgfile == '\0') dbgfile = av[++ax];
		}
		else if (strcmp(arg, "--debug") == 0) {
		    debug = YES;
			dbgfile = av[++ax];
		}
		
		/* -Ln or -L n or --ciflevel n */ 
		NUMERICARG("-L","--ciflevel",ciflevel,"CIF level",0,1)                                \

        /* -lnnnn or -l nnnn or --maxline nnnn */
		NUMERICARG("-l","--maxline",maxline,"maximum line length",1,2048)
		
		/* -s or --fullstar */
		else if (strcmp(arg, "-s") == 0 || strcmp(arg, "--fullstar") == 0 ) {
			full_star  = YES;
		}
		
		/* -b or --printdbn */
		else if (strcmp(arg, "-b") == 0 || strcmp(arg, "--printdbn") == 0) {
			reportlevel = QUIET;
			printdblkname = YES;
		}
		
		/* -q or --quiet */
		else if (strcmp(arg, "-q") == 0 || strcmp(arg, "--quiet") == 0) {
			reportlevel = QUIET;
		}
		
		/* -v or --verbose */
		else if (strcmp(arg, "-v") == 0 || strcmp(arg, "--verbose") == 0) {
			reportlevel = VERBOSE;
		}
		
		/* -ennnn or -e nnnn or --errors nnnn */
		NUMERICARG("-e","--errors",error_threshold,"errors to report",0,4096)

        /* -wnn or -w nn or --maxdname nn */
		NUMERICARG("-w","--maxdname",maxdname,"maximum width of data name",1,2043)

        /* -h or --help */
		else if (strcmp(arg, "-h") == 0 || strcmp(arg, "--help") == 0) {
			usage();
			return 0;
		}
		
		/* -V or --version */
		else if (strcmp(arg, "-V") == 0) {
			fprintf(stderr, "%s version %s\n", av[0], VERSION);
			return 0;
		}
		else if (strcmp(arg, "--") == 0) {
		    ignoreargs = YES;
		}
		else if (ax < (ac-1)) {
			errflg++;
			break;
		} 
		else {
		  inputfile = arg;
		}
	}
		if (errflg || ax != ac) 
		{
			usage();
			exit(2);
		}

		if (ax == ac-1) {
			inputfile = av[ax];
		}
		

        /* clean up interaction among ciflevel, maxline and maxdname */

       if ( ciflevel == 0 ) {
	      /* The user has selected CIF 1.0 */
		  if ( maxdname < 0 ) maxdname = CIFMAXDNAME-1;
		  if ( maxline < 0 ) maxline = 80;
	   }
	   else if (ciflevel == 1 || ciflevel == -1 ) {
	      /* The user has selected CIF 1.1 */
		  ciflevel = 1;
		  if ( maxdname < 0 ) maxdname = CIFMAXDNAME;
		  if ( maxline < 0 ) maxline = 2048;
	   } else {
	      fprintf(stderr, "vcif ERROR: invalid CIF level\n");
		  usage();
		  exit(2);
	   }
	   
	   if (maxdname > maxline-5) {
	     fprintf(stderr, "vcif ERROR: maxdname > maxline-5\n");
		 usage();
		 exit(2);
	   }
	   
/* handle user-specified error threshold */
       if ( error_threshold > 0 ) force_threshold = YES;
       if ( error_threshold < 0 ) error_threshold = 20;

/* If no arguments, or if input filename is given as "-", accept stdin input */
	if (strcmp(inputfile, "-") == 0 || strcmp(inputfile, "stdin") == 0) {
		cif = stdin;
	}
	else
/* Must be able to open the file for checking! */
	if ((cif = fopen(inputfile, "r")) == NULL) {
		fprintf(stderr, "vcif ERROR: unable to open file %s - exiting\n", inputfile);
		exit(2);
	}
	
	pass1(cif);
	if (errors_reported == 0) {
		load_dictionary_info();
		pass2();
	}

	fclose(cif);

	if (errors_reported < MAXERRS ) {
		return errors_reported;
	} else {
		return MAXERRS - 1;
	}

}	/* MAIN */


void load_dictionary_info(){	/* stub for later development */
	return;
}

void pass2(){			/* stub for later development */
	return;
}

void vcifwarn(message, note)
	char * message;
	int note;
{

	int i;			/* scratch integer */
	static int lastlineno = 0;		/* remembers last line number */

	if (reportlevel == NORMAL) {
		if (force_threshold == NO) {
		  fprintf(stdout, "%s", message);
		}
		else {
			if (errors_reported < error_threshold) {
				fprintf(stdout, "%s", message);
			}
			if (errors_reported == error_threshold) {
				fprintf(stdout, "%s", "-------------------------------------------------------------------------\n");
				fprintf(stdout, "Only the first %d errors are printed out.\n", error_threshold);
			}
		}
	}

	if (reportlevel == VERBOSE) {
		if (errors_reported < error_threshold) {
			fprintf(stdout, "%s", message);
			if (lineno != lastlineno) {
				fprintf(stdout, ">>> \"");
				for (i=0; i < strlen(buffer); i++) {
			 		if (i > 50) {
						if (i<54) fprintf(stdout, ".");
					}
			 		else fprintf(stdout, "%c", buffer[i]);
			 		}
				fprintf(stdout, "\"\n");
			}
			lastlineno = lineno;
			elaborate(note);
		}

		if (errors_reported == error_threshold) {
			elaborate(note);
			fprintf(stdout, "-------------------------------------------------------------------------\n");
			fprintf(stdout, "Only the first %d errors are printed out. Please fix these and try again!\n", error_threshold);
			fprintf(stdout, "-------------------------------------------------------------------------\n");
		}
	}
	errors_reported++;
	return;
}

/*
______________________________________________________________________________

usage() prints out the usage message on stderr
______________________________________________________________________________
*/

int usage()
{
	fprintf(stderr, "Usage: vcif [options] file \n");
	fprintf(stderr, "       vcif -help | -version \n");
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "       -L n              or\n");
	fprintf(stderr, "       -ciflevel n       n = 0 for CIF 1.0, 1 for CIF 1.1\n");
	fprintf(stderr, "       --maxline nnnn maximum line length\n");
	
	fprintf(stderr, "       -e nn          no. errors reported\n");
	fprintf(stderr, "       -h             print this usage info\n");
	fprintf(stderr, "       -l nnnn           or\n");
	fprintf(stderr, "       --maxline nnnn    maximum line length\n");
	fprintf(stderr, "       -q                or\n");
	fprintf(stderr, "       --quiet           quiet: report no errors\n");
	fprintf(stderr, "       -b                or \n");
	fprintf(stderr, "       --printdbn        quiet: print datablock names\n");
	fprintf(stderr, "       -s                or \n");
	fprintf(stderr, "       -fullstar         allow extra STAR rules\n");
	fprintf(stderr, "       -v                or\n");
	fprintf(stderr, "       --verbose         verbose: give extra info\n");
	fprintf(stderr, "       -w nn             or\n");
	fprintf(stderr, "       --maxdname nn     maximum dataname length\n");
	fprintf(stderr, "       -D filename       or\n");
	fprintf(stderr, "       --debug filename  file for debugging info\n");
	fprintf(stderr, "       -V                or\n");
	fprintf(stderr, "       --version         print version number\n");

	exit(1);
}	

int elaborate(note)
	int note;
{
#define NSIZE 32
	static int printed[NSIZE] = {
	NO, NO, NO, NO, NO, NO, NO, NO,
	NO, NO, NO, NO, NO, NO, NO, NO,
	NO, NO, NO, NO, NO, NO, NO, NO,
	NO, NO, NO, NO, NO, NO, NO, NO
	};

	static char *explanation[NSIZE] = {

/* 0 */
"",
/* 1 */
/*0 "Error! unable to open debug file - exiting"
0 "WARNING: line %d: length exceeds internal buffer size (%d) - rest discarded"
0 "WARNING: Line %d exceeds %d characters"*/
"*** The indicated line is longer than the length you specified with -l",
/* 2 */
/* 1 "WARNING: Line %d: dataname exceeds %d characters (%s)"
1 "WARNING: Line %d: dataname exceeds %d characters (%s)"*/
"*** The indicated dataname is longer than the length you specified with -w",
/* 3 */
/* 1 "CIF ERROR: Line %d exceeds %d characters" */
"*** The indicated line is longer than the maximum 80 characters permitted in\n    the CIF specification.",
/* 4 */
/* 1 "CIF ERROR: Line %d: dataname exceeds %d characters (%s)" */
"*** A dataname (beginning with an underscore _) or a data block identifier\n    (beginning data_) is longer than the maximum length of 74 characters\n    consistent with the 80-character line length limit.",
/* 5 */
/* "Double-quoted character string does not terminate at line %d"*/
"*** The indicated line appears to contain some word or words introduced by\n    double quotes, but not terminated with matching double quotes.",
/* 6 */
/* 1 "Single-quoted character string does not terminate at line %d"*/
"*** The indicated line appears to contain some word or words introduced by\n    a single quote, but not terminated with a matching single quote.",
/* 7 */
/* 1 "ERROR: Missing data value before data block code at line %d"*/
"*** The last dataname in a data block appears not to have an associated value.",
/* 8 */
/* 1 "ERROR: No data block code before dataname at line %d" */
"*** A data block MUST begin with a data_something declaration.",
/* 9 */
/* 1 "ERROR: Redundant loop_ keyword at line %d"*/
"*** Two or more loop_ keywords appear together. Nested loops aren't permitted\n    in CIFs.",
/* 10 */
/* 1 "ERROR: Unexpected string (%s) at line %d"
1 "Unexpected string <%s> at line %d"
1 "Unexpected text field beginning at line %d" */
"*** There is an unexpected word or number as indicated. This may be because\n    a loop is intended but the loop_ keyword has been missed out; or a phrase\n    with several words is not enclosed in matching delimiting quote marks;\n    or a text field (extending over several lines) is not properly closed\n    with a final semicolon; or a data_ block header has not yet been seen.",
/* 11 */
/* 2 "ERROR: missing data value at end of file"
1 "ERROR: Value missing at line %d" */
"*** A dataname is not accompanied by a matching value. The dataname may be\n    on the line preceding the one indicated.",
/* 12 */
/* 1 "ERROR: data_ declaration at line %d needs identifier" */
"*** The data_ statement which separates data blocks must have an identifying\n    string, i.e. say data_something, where the something is an arbitrary label\n    (but must be unique within the file).",
/* 13 */
/* 1 "ERROR: duplicate datablock name data_%s at line %d" */
"*** This file contained another data_ block header with the same label.",
/* 14 */
/* 1 "ERROR: duplicate dataname %s in data_%s at line %d"*/
"*** A dataname must occur once only in any data block.",
/* 15 */
/* 2 "ERROR: loop takes no data values before end of file"
1 "ERROR: loop takes no data values before data block code at line %d" */
"*** A loop structure has been declared, but there are no values in it.",
/* 16 */
/* 1 "ERROR: loop_ keyword out of context at line %d" */
"*** The loop_ keyword was not expected here. Perhaps there is a missing\n    data value.",
/* 17 */
/* 1 "ERROR: null datablock at end of file"
1 "ERROR: null datablock prior to line %d" */
"*** A data_ block header has occurred, but there are no data names or data\n    values in the data block thus set up.",
/* 18 */
/* 1 "ERROR: single underscore ('_') at line %d not defined" */
"*** The underscore character by itself has no meaning in a CIF. Has it come\n    adrift from an intended dataname?",
/* 19 */
/* 1 "File begins incorrectly with text (%s) at line %d" */
"*** The CIF must begin with a data_ block header and contain datanames and\n    data values in conformance with the CIF specification.",
/* 20 */
/* 1 "WARNING: non-printable ASCII character \\%#o at line %d" */
"*** The CIF should contain only alphanumeric or punctuation characters\n    defined as printable in the ASCII character set.",
/* 21 */
/* 1 "ERROR: Number of loop elements not multiple of packetsize at line %d" */
"*** A loop_ header defines a list of datanames. The values following this\n    header are assigned in sequence with the datanames in the header, so\n    each packet of information (or row in the table of values defined by the\n    loop structure) must have the same number of values as there are datanames\n    declared in the loop header. Common reasons for this error include:\n    omission of a value where the associated data are absent (insert . or ?\n    as placeholders); numeric values where the standard uncertainty (or e.s.d)\n    has come adrift from its associated value (e.g. 10.925 (2)); multi-word\n    phrases or text entries that are not properly delimited with quote marks\n    or initial semicolons.\n!   NOTE: the problem may lie in the loop prior to the indicated line number.",
/* 22 */
/* 2 "ERROR: quoted character string at end of file does not terminate"
2 "ERROR: text field at end of file does not terminate" */
"*** Is the CIF complete?",
/* 23 */
/* 2 "WARNING: null file"*/
"*** No valid data_ block identifier or dataname was found: the file was\n    empty or contained only comment lines.",
/* 24 */
"*** A loop_ keyword was seen, indicating the start of a loop header, but\n    no dataname(s) followed before a value was seen.",
/* 25 */ "",
/* 26 */ "",
/* 27 */ "",
/* 28 */ "",
/* 29 */ "",
/* 30 */ "",
/* 31 */ ""
	};

	if (note == 0) return 0;

	if (printed[note] == NO && (errors_reported < (error_threshold + 1))) {
		printed[note] = YES;
		fprintf(stdout, "%s\n", explanation[note]);
	}
	return 0;
}
